# -*- coding: utf-8 -*-
import os
import uuid
import redis
import functools
from abc import ABCMeta, abstractmethod

try:
    import cPickle as pickle
except ImportError:
    import pickle as pickle


def same_types(fn):
    @functools.wraps(fn)
    def wrapper(self, *args):
        types = (self.__class__,) + self._same_types

        allowed = all([
            any([isinstance(arg, t) for t in types])
            for arg in args
        ])

        if not allowed:
            types_msg = ', '.join(types[:-1])
            types_msg = ' or '.join([types_msg, types[-1]])
            message = ('Only instances of %s are '
                       'supported as operand types.') % types_msg
            raise TypeError(message)

        return fn(self, *args)
    return wrapper


class HoneyBean:
    """
        蜜豆, 抽象类为其它数据结构提供后端功能
    """

    __metaclass__ = ABCMeta
    _same_types = ()

    @abstractmethod
    def __init__(self, data=None, redis=None, key=None, pickler=None):
        self.redis = redis or self._create_redis()
        self.pickler = pickler or pickle
        self.key = key or self._create_key()

        # 初始化
        if data is not None:
            if isinstance(data, HoneyBean):
                # 为事务做一层包装
                def init_trans(pipe):
                    d = data._data(pipe=pipe)
                    pipe.multi()
                    self._init_data(d, pipe=pipe)
                self._transaction(init_trans)
            else:
                self._init_data(data)

    def _create_new(self, data=None, key=None, pipe=None, cls=None):
        assert not isinstance(data, HoneyBean), \
            u"非事务操作，不能保证原子性. 使用_data() 开始一个数据事务"

        cls = cls or self.__class__
        if issubclass(cls, HoneyBean):
            settings = {
                'key': key,
                'redis': self.redis,
                'pickler': self.pickler,
            }

            if pipe is not None and data:
                new = cls(**settings)
                new._init_data(data, pipe=pipe)
                return new

            return cls(data, **settings)

        return cls(data) if data else cls()

    def _init_data(self, data, pipe=None):
        assert not isinstance(data, HoneyBean), \
            u"非事务操作，不能保证原子性. 使用_data() 开始一个数据事务"

        # 提供使用外部redis pipe实例的可能参数
        p = pipe if pipe is not None else self.redis.pipeline()
        if data is not None:
            self._clear(pipe=p)
            if data:
                self._update(data, pipe=p)

        if pipe is None:
            p.execute()

    def _create_redis(self):
        """
        默认初始化redis 实例通过honey.conf的配置 class:`redis.StrictRedis`
        """
        import yaml

        redis_conf = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'conf', 'honey.conf')

        if not os.path.exists(redis_conf):
            redis_conf = os.path.join(os.path.expanduser('~'), '.honey.conf')

        with open(redis_conf, 'r') as r:
            conf = yaml.load(r)
            if 'redis' in conf:
                return redis.StrictRedis(**conf['redis'])

        return redis.StrictRedis()

    def _create_key(self):
        """
        创建一个redis 的存储key
        :rtype: string

        这里使用uuid4 生成长度32位的key. 发生key有效碰撞的概率接近0
        参考帖子 http://stackoverflow.com/a/786541/325365
        """
        return uuid.uuid4().hex

    @abstractmethod
    def _data(self, pipe=None):
        """
        内部一个事务的开始
        """
        pass

    def _pickle(self, data):
        return str(self.pickler.dumps(data))

    def _unpickle(self, string):
        if string is None:
            return None
        if not isinstance(string, basestring):
            msg = 'Only strings can be unpickled (%r given).' % string
            raise TypeError(msg)

        return self.pickler.loads(string)

    @abstractmethod
    def _update(self, data, pipe=None):
        assert not isinstance(data, HoneyBean), \
            u"非事务操作，不能保证原子性. 使用_data() 开始一个数据事务"

    def _clear(self, pipe=None):
        redis = pipe if pipe is not None else self.redis
        redis.delete(self.key)

    def clear(self):
        self._clear()

    def _transaction(self, fn, *extra_keys):
        results = []

        def trans(pipe):
            results.append(fn(pipe))

        self.redis.transaction(trans, self.key, *extra_keys)
        return results[0]

    def copy(self, key=None):
        def copy_trans(pipe):
            data = self._data(pipe=pipe)
            pipe.multi()
            return self._create_new(data, key=key, pipe=pipe)
        return self._transaction(copy_trans, key)

    def _repr_data(self, data):
        return repr(data)

    def __repr__(self):
        cls_name = self.__class__.__name__
        data = self._repr_data(self._data())
        return '<redis_collections.%s at %s %s>' % (cls_name, self.key, data)



# -*- coding: utf-8 -*-
"""
蜂巢的hash 结构接口，一般用于缓存或关系存储.
"""

import collections
from .base import HoneyBean, same_types


class Dict(HoneyBean, collections.MutableMapping):

    """
        缺失值的提示处理, 遵守MutableMapping 抽象基类的定义
    """
    class __missing_value(object):
        def __repr__(self):
            return '<missing value>'

    __marker = __missing_value()

    def __init__(self, *args, **kwargs):
        """
        初始化Base的参数
        :param data: Initial data.
        :type data: iterable or mapping
        :param redis: Redis client instance. If not provided, default Redis
                      connection is used.
        :type redis: :class:`redis.StrictRedis`
        :type key: str
        :param pickler: 数据的自定义序列化参数 Examples::

                            import json, pickle
                            Dict(pickler=json)
                            Dict(pickler=pickle)  # default
        """
        super(Dict, self).__init__(*args, **kwargs)

    def __len__(self):
        """返回对应Key的hash字典里一共有多少个item，like: len()"""
        return self.redis.hlen(self.key)

    def __iter__(self):
        """返回一个字典keys的迭代器."""
        return iter(self.redis.hkeys(self.key))

    def __contains__(self, key):
        """
        指定的key是否存在，返回boolean.
        """
        return self.redis.hexists(self.key, key)

    def get(self, key, default=None):
        """
        和Python的字典一样的用法，为空的话，提供默认值填充
        """
        value = self.redis.hget(self.key, key)
        return self._unpickle(value) or default

    def getmany(self, *keys):
        """
        给定一个key的list，批量返回这些key的值.
        """
        values = self.redis.hmget(self.key, *keys)
        return map(self._unpickle, values)

    def __getitem__(self, key):
        """
        同Python的dict，通过``d[key]``方式返回对应的值，若不存在抛出KeyError
        """
        with self.redis.pipeline() as pipe:
            pipe.hexists(self.key, key)
            pipe.hget(self.key, key)
            exists, value = pipe.execute()

        if not exists:
            if hasattr(self, '__missing__'):
                return self.__missing__(key)
            raise KeyError(key)
        return self._unpickle(value)

    def __setitem__(self, key, value):
        """ 字典的保存方式``d[key]`` to *value*."""
        value = self._pickle(value)
        self.redis.hset(self.key, key, value)

    def __delitem__(self, key):
        """
        移除一个缓存值，字典的方式 ``del d[key]``
        """
        with self.redis.pipeline() as pipe:
            pipe.hexists(self.key, key)
            pipe.hdel(self.key, key)
            exists, _ = pipe.execute()

        if not exists:
            raise KeyError(key)

    def discard(self, key):
        """
        Redis的方式移除一个值. 简单直接 Remove ``d[key]``
        """
        self.redis.hdel(self.key, key)

    def _data(self, pipe=None):
        """
        返回全部以hash方式存储的items ``(key, value)`` pairs 列表. 一般用 items() 替代直接访问
        """
        redis = pipe if pipe is not None else self.redis
        result = redis.hgetall(self.key).items()
        return [(k, self._unpickle(v)) for (k, v) in result]

    def items(self):
        """替代直接 _data. 推荐iteritems"""
        return self._data()

    def iteritems(self):
        """返回一个可迭代的键值对 ``(key, value)`` pairs."""
        result = self.redis.hgetall(self.key).iteritems()
        return ((k, self._unpickle(v)) for (k, v) in result)

    def keys(self):
        """返回字典实例key下的所有keys. 比如以项目为主key d['feed'] -> [1,2,3,4,5] """
        return self.redis.hkeys(self.key)

    def iter(self):
        """Key的迭代器"""
        return self.__iter__()

    def iterkeys(self):
        """Python dictionary的实现."""
        return self.__iter__()

    def values(self):
        """返回全部的values."""
        result = self.redis.hvals(self.key)
        return [self._unpickle(v) for v in result]

    def itervalues(self):
        """values 的可迭代版本."""
        result = iter(self.redis.hvals(self.key))
        return (self._unpickle(v) for v in result)

    def pop(self, key, default=__marker):
        """
        和Python dict 的pop实现功能一致，返回对应key的value，同时移除
        """
        with self.redis.pipeline() as pipe:
            pipe.hget(self.key, key)
            pipe.hdel(self.key, key)
            value, existed = pipe.execute()

        if not existed:
            if default is self.__marker:
                raise KeyError(key)
            return default
        return self._unpickle(value)

    def popitem(self):
        """
        返回并移除一个item
        """
        def popitem_trans(pipe):
            try:
                key = pipe.hkeys(self.key)[0]
            except IndexError:
                raise KeyError

            pipe.multi()
            pipe.hget(self.key, key)
            pipe.hdel(self.key, key)
            value, _ = pipe.execute()

            return key, value

        key, value = self._transaction(popitem_trans)
        return key, self._unpickle(value)

    def setdefault(self, key, default=None):
        """
        和Python dictionary的setdefault一致，可用于嵌套字典的存储
        """
        with self.redis.pipeline() as pipe:
            pipe.hsetnx(self.key, key, self._pickle(default))
            pipe.hget(self.key, key)
            _, value = pipe.execute()
        return self._unpickle(value)

    def _update(self, data, pipe=None):
        super(Dict, self)._update(data, pipe)
        redis = pipe if pipe is not None else self.redis

        data = dict(data)
        keys = data.keys()
        values = map(self._pickle, data.values())

        redis.hmset(self.key, dict(zip(keys, values)))

    def update(self, other=None, **kwargs):
        """
        把已有的Hash实例的字典更新进来
        使用方式也可以这样: ``d.update(a=1, b=2)``.
        """
        other = other or {}
        if isinstance(other, HoneyBean):
            def update_trans(pipe):
                d = other._data(pipe=pipe)
                pipe.multi()
                self._update(d, pipe=pipe)
            self._transaction(update_trans)
        else:
            mapping = {}
            mapping.update(other, **kwargs)
            self._update(mapping)

    @classmethod
    def fromkeys(cls, seq, value=None, **kwargs):
        """
        创建一个新的Hash 字典存储实例，参数的seq 可以是list，tuple等，默认创建的存储字典的value为None
        """
        values = ((item, value) for item in seq)
        return cls(values, **kwargs)

    def _repr_data(self, data):
        return repr(dict(data))


class Counter(Dict):
    """
    以字典的方式去做计数
    """

    _same_types = (collections.Counter,)

    def __init__(self, *args, **kwargs):
        """
        计数类. 应为业务简单，只返回数值，所以不支持复杂存储结构，也不支持pickle.
        :param data: 初始数据 ``key value`` pair.

        初始化语法: ``c = Counter(a=1, b=2)``
        """
        if 'pickler' in kwargs:
            del kwargs['pickler']
        super(Counter, self).__init__(*args, **kwargs)

    def _pickle(self, data):
        return unicode(int(data))

    def _unpickle(self, string):
        if string is None:
            return None
        return int(string)

    def _obj_to_data(self, obj):
        is_redis = isinstance(obj, HoneyBean)
        is_mapping = isinstance(obj, collections.Mapping)

        data = obj._data() if is_redis else obj
        return dict(data) if is_mapping else iter(data)

    def getmany(self, *keys):
        """
        这里注意，默认值都为0
        """
        values = super(Counter, self).getmany(*keys)
        return [(v or 0) for v in values]

    def __getitem__(self, key):
        return self.get(key, 0)

    def elements(self):
        """
        一个迭代器. 返回计数元素.
        """
        for element, count in self._data():
            if count:
                for _ in xrange(0, count):
                    yield element

    def _update(self, data, pipe=None):
        super(Dict, self)._update(data, pipe)
        redis = pipe if pipe is not None else self.redis

        data = collections.Counter(data)
        keys = data.keys()
        values = map(self._pickle, data.values())

        redis.hmset(self.key, dict(zip(keys, values)))

    def incr(self, key, n=1):
        """
        自增方法. 指定计数key，默认自增1
        :rtype: integer
        """
        if n:
            return self.redis.hincrby(self.key, key, self._pickle(n))
        return 0

    def decr(self, key, n=-1):
        """
        本不应该提供单独的自减函数，包一个算了
        """
        return self.incr(key, n=n)

    def most_common(self, n=None):
        """
        实现类collections Counter的方法. 返回最多的Top n计数元素
        """
        data = self._obj_to_data(self)
        counter = collections.Counter(data)
        return counter.most_common(n)

    def _operation(self, other, fn, update=False):
        """
        内部操作实现函数
        """
        key = self.key if update else None

        def op_trans(pipe):
            d1 = self._obj_to_data(self)
            d2 = self._obj_to_data(other)

            c1 = collections.Counter(d1)
            result = fn(c1, d2)

            if update:
                result = c1

            pipe.multi()
            return self._create_new(result, key=key, pipe=pipe)
        return self._transaction(op_trans, key)

    def subtract(self, other):
        """
        使用另一个Counter对象更新，对应key的计数相减
        """
        def subtract_op(c1, d2):
            c1.subtract(d2)
        self._operation(other, subtract_op, update=True)

    def update(self, other):
        """
        将另一个Counter对象更新到当前Counter
        """
        def update_op(c1, d2):
            c1.update(d2)
        self._operation(other, update_op, update=True)

    @same_types
    def __add__(self, other):
        def add_op(c1, d2):
            c2 = collections.Counter(d2)
            return c1 + c2
        return self._operation(other, add_op)

    def __radd__(self, other):
        return self.__add__(other)

    @same_types
    def __iadd__(self, other):
        def iadd_op(c1, d2):
            c2 = collections.Counter(d2)
            c1 += c2
        return self._operation(other, iadd_op, update=True)

    @same_types
    def __and__(self, other):
        def and_op(c1, d2):
            c2 = collections.Counter(d2)
            return c1 & c2
        return self._operation(other, and_op)

    def __rand__(self, other):
        return self.__and__(other)

    @same_types
    def __iand__(self, other):
        def iand_op(c1, d2):
            c2 = collections.Counter(d2)
            c1 &= c2
        return self._operation(other, iand_op, update=True)

    @same_types
    def __sub__(self, other):
        def sub_op(c1, d2):
            c2 = collections.Counter(d2)
            return c1 - c2
        return self._operation(other, sub_op)

    @same_types
    def __rsub__(self, other):
        def rsub_op(c1, d2):
            c2 = collections.Counter(d2)
            return c2 - c1
        return self._operation(other, rsub_op)

    @same_types
    def __isub__(self, other):
        def isub_op(c1, d2):
            c2 = collections.Counter(d2)
            c1 -= c2
        return self._operation(other, isub_op, update=True)

    @same_types
    def __or__(self, other):
        def or_op(c1, d2):
            c2 = collections.Counter(d2)
            return c1 | c2
        return self._operation(other, or_op)

    def __ror__(self, other):
        return self.__or__(other)

    @same_types
    def __ior__(self, other):
        def ior_op(c1, d2):
            c2 = collections.Counter(d2)
            c1 |= c2
        return self._operation(other, ior_op, update=True)

    @classmethod
    def fromkeys(cls, seq, value=None, **kwargs):
        raise NotImplementedError



# -*- coding: utf-8 -*-
"""
蜂巢的List 数据结构接口
"""
import collections
from .base import HoneyBean


class List(HoneyBean, collections.MutableSequence):

    def __init__(self, *args, **kwargs):
        super(List, self).__init__(*args, **kwargs)

    def __len__(self):
        return self.redis.llen(self.key)

    def _data(self, pipe=None):
        redis = pipe if pipe is not None else self.redis
        values = redis.lrange(self.key, 0, -1)
        return (self._unpickle(v) for v in values)

    def __iter__(self):
        return self._data()

    def __reversed__(self):
        values = self.redis.lrange(self.key, 0, -1)
        return (self._unpickle(v) for v in reversed(values))

    def _recalc_slice(self, start, stop):
        start = start or 0
        if stop is None:
            stop = -1
        else:
            stop = stop - 1 if stop != 0 else stop
        return start, stop

    def _calc_overflow(self, size, index):
        return (index >= size) if (index >= 0) else (abs(index) > size)

    def _get_slice(self, index):
        assert isinstance(index, slice)

        def slice_trans(pipe):
            start, stop = self._recalc_slice(index.start, index.stop)
            values = pipe.lrange(self.key, start, stop)
            if index.step:
                # 实现python 切片的技巧
                values = values[::index.step]
            values = map(self._unpickle, values)

            pipe.multi()
            return self._create_new(values, pipe=pipe)
        return self._transaction(slice_trans)

    def __getitem__(self, index):
        if isinstance(index, slice):
            return self._get_slice(index)

        with self.redis.pipeline() as pipe:
            pipe.llen(self.key)
            pipe.lindex(self.key, index)
            size, value = pipe.execute()

        if self._calc_overflow(size, index):
            raise IndexError(index)
        return self._unpickle(value)

    def get(self, index, default=None):
        value = self.redis.lindex(self.key, index)
        return self._unpickle(value) or default

    def _set_slice(self, index, value):
        assert isinstance(index, slice)

        if value:
            raise NotImplementedError(self.not_impl_msg)
        self.__delitem__(index)

    def __setitem__(self, index, value):
        if isinstance(index, slice):
            self._set_slice(index, value)
        else:
            def set_trans(pipe):
                size = pipe.llen(self.key)
                if self._calc_overflow(size, index):
                    raise IndexError(index)
                pipe.multi()
                pipe.lset(self.key, index, self._pickle(value))

            self._transaction(set_trans)

    def _del_slice(self, index):
        assert isinstance(index, slice)

        begin = 0
        end = -1

        if index.step:
            raise NotImplementedError(self.not_impl_msg)

        start, stop = self._recalc_slice(index.start, index.stop)

        if start == begin and stop == end:
            self.clear()
            return

        with self.redis.pipeline() as pipe:
            if start != begin and stop == end:
                pipe.ltrim(self.key, begin, start - 1)
            elif start == begin and stop != end:
                pipe.ltrim(self.key, stop + 1, end)
            else:
                raise NotImplementedError(self.not_impl_msg)
            pipe.execute()

    def __delitem__(self, index):
        begin = 0
        end = -1

        if isinstance(index, slice):
            self._del_slice(index)
        else:
            if index == begin:
                self.redis.lpop(self.key)
            elif index == end:
                self.redis.rpop(self.key)
            else:
                raise NotImplementedError(self.not_impl_msg)

    def remove(self, value):
        self.redis.lrem(self.key, 1, self._pickle(value))

    def index(self, value, start=None, stop=None):
        start, stop = self._recalc_slice(start, stop)
        values = self.redis.lrange(self.key, start, stop)

        for i, v in enumerate(self._unpickle(v) for v in values):
            if v == value:
                return i + start
        raise ValueError(value)

    def count(self, value):
        return list(self._data()).count(value)

    def insert(self, index, value):
        if index != 0:
            raise NotImplementedError(self.not_impl_msg)

        self.redis.lpush(self.key, self._pickle(value))

    def append(self, value):
        self.redis.rpush(self.key, self._pickle(value))

    def _update(self, data, pipe=None):
        super(List, self)._update(data, pipe)
        redis = pipe if pipe is not None else self.redis

        values = map(self._pickle, data)
        redis.rpush(self.key, *values)

    def extend(self, values):
        if isinstance(values, HoneyBean):
            def extend_trans(pipe):
                d = values._data(pipe=pipe)  # retrieve
                pipe.multi()
                self._update(d, pipe=pipe)  # store
            self._transaction(extend_trans)
        else:
            self._update(values)

    def pop(self, index=-1):
        if index == 0:
            value = self.redis.lpop(self.key)
        elif index == -1:
            value = self.redis.rpop(self.key)
        else:
            raise NotImplementedError(self.not_impl_msg)
        return self._unpickle(value)

    def __add__(self, values):
        def add_trans(pipe):
            d1 = list(self._data(pipe=pipe))  # retrieve

            if isinstance(values, HoneyBean):
                d2 = list(values._data(pipe=pipe))  # retrieve
            else:
                d2 = list(values)

            pipe.multi()
            return self._create_new(d1 + d2, pipe=pipe)  # store
        return self._transaction(add_trans)

    def __radd__(self, values):
        return self.__add__(values)

    def __mul__(self, n):
        if not isinstance(n, int):
            raise TypeError('Cannot multiply sequence by non-int.')

        def mul_trans(pipe):
            data = list(self._data(pipe=pipe))  # retrieve
            pipe.multi()
            return self._create_new(data * n, pipe=pipe)  # store
        return self._transaction(mul_trans)

    def __rmul__(self, n):
        return self.__mul__(n)

    def _repr_data(self, data):
        return repr(list(data))
